/*
  PostStoryDiff.cpp - Increase some NPC encounter toughness after the story.

  Jason Hood, 22 May, 2023.

  NPCs are spawned according to the difficulty set in the encounter entry of a
  zone.  This plugin will increase that difficulty after the story.  If the
  highest difficulty in the system is less than 12, add the difference to all
  the encounters.  E.g. LI01 has encounters ranging from 1 to 4; 12 - 4 = 8;
  add 8 to all the encounters, thus becoming 9 to 12.

  Install:
	Copy PostStoryDiff.DLL to your DLLS\BIN directory and add it to the
	[Initial SP DLLs] section of EXE\freelancer.ini.
*/

#include "Common.h"

#define NAKED	__declspec(naked)
#define STDCALL __stdcall
#define EXPORT	__declspec(dllexport)


#define ADDR_DIFF11 ((PDWORD)(content + 0x6ef9066+1 - 0x6ea0000))
#define ADDR_READ11 ((PDWORD)(content + 0x6f74b78+1 - 0x6ea0000))
#define ADDR_MNUM11 ((PDWORD)(content + 0x6ea6a7f+1 - 0x6ea0000))

#define ADDR_DIFF10 ((PDWORD)(content + 0x6fe92d6+1 - 0x6f90000))
#define ADDR_READ10 ((PDWORD)(content + 0x7064a48+1 - 0x6f90000))
#define ADDR_MNUM10 ((PDWORD)(content + 0x6f96a4f+1 - 0x6f90000))


DWORD dummy;
#define ProtectX( addr, size ) \
  VirtualProtect( addr, size, PAGE_EXECUTE_READWRITE, &dummy )

#define NEWOFS( from, to, org ) \
  org = (DWORD)(from) + *(PDWORD)(from) + 4; \
  *(PDWORD)(from) = (DWORD)(to) - (DWORD)(from) - 4


bool post_story;
int  diff_adjust, last_mnum;


void STDCALL FindEncounters( INI_Reader& ini )
{
  diff_adjust = 0;
  if (post_story)
  {
    int maxdiff = 0;
    while (ini.read_header())
    {
      if (ini.is_header( "zone" ))
      {
	while (ini.read_value())
	{
	  if (ini.is_value( "encounter" ))
	  {
	    int diff = ini.get_value_int( 1 );
	    if (diff > maxdiff)
	      maxdiff = diff;
	  }
	}
      }
    }
    ini.reset();
    if (maxdiff < 12)
      diff_adjust = 12 - maxdiff;
  }
}


DWORD Diff_Org, Read_Org, MNum_Org;


NAKED
void Diff_Hook()
{
  __asm {
	call	Diff_Org		// ftol
	add	eax, diff_adjust
	ret
  }
}


NAKED
void Read_Hook()
{
  __asm {
	push	ecx
	push	ebx
	call	FindEncounters
	pop	ecx
	jmp	Read_Org
	align	16
  }
}


NAKED
void MNum_Hook()
{
  __asm {
	cmp	eax, last_mnum
	sete	post_story
	jmp	MNum_Org
	align	16
  }
}


struct IDLL
{
  virtual void whatever( int, LPVOID ) { }
};


extern "C" EXPORT
LPVOID CreateInstance( LPVOID, LPCSTR, LPVOID )
{
  PBYTE content = (PBYTE)GetModuleHandle( "content.dll" );
  if (*((PBYTE)ADDR_DIFF11-1) == 0xE8)
  {
    ProtectX( ADDR_DIFF11, 4 );
    ProtectX( ADDR_READ11, 4 );
    ProtectX( ADDR_MNUM11, 4 );
    NEWOFS( ADDR_DIFF11, Diff_Hook, Diff_Org );
    NEWOFS( ADDR_READ11, Read_Hook, Read_Org );
    NEWOFS( ADDR_MNUM11, MNum_Hook, MNum_Org );
    last_mnum = 42;
  }
  else
  {
    ProtectX( ADDR_DIFF10, 4 );
    ProtectX( ADDR_READ10, 4 );
    ProtectX( ADDR_MNUM10, 4 );
    NEWOFS( ADDR_DIFF10, Diff_Hook, Diff_Org );
    NEWOFS( ADDR_READ10, Read_Hook, Read_Org );
    NEWOFS( ADDR_MNUM10, MNum_Hook, MNum_Org );
    last_mnum = 41;
  }

  return new IDLL;
}


extern "C" EXPORT
void DestroyInstance( IDLL& )
{
  return;
}
